/* 
 * absyn.c
 * This file will define all functions
 * for building the abstract syntax tree
 * Please note that the abstract syntax tree is not an exact instance
 * of the abstract syntax defined in the OCL2.0 specification, it is 
 * just a syntax tree build from the concrete syntax by discarding 
 * the unnecessary info
 */

#include "util.h"
#include "absyn.h"
#include "creat_python.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


string my_default_type = "DEFAULT";

static float coord_y = Y_INIT;

extern int obj_ind;


/*************************************************************************
 * Constructors
 *************************************************************************/
A_oclExpression
A_OclExpression(void *expr,NodeType type)
{
  A_oclExpression e = (A_oclExpression)checked_malloc(sizeof(*e));
  e->type = OCLEXPRESSION;
  switch (type)
    {
    case PROPERTYCALLEXP:
      e->kind = A_PROPERTYCALLEXP;
      e->u.propertyCallExp = (A_propertyCallExp)expr;
      return e;
    case VARIABLEEXP:
      e->kind = A_VARIABLEEXP;
      e->u.variableExp = (A_variableExp)expr;
      return e;
    case LITERALEXP:
      e->kind = A_LITERALEXP;
      e->u.literalExp = (A_literalExp)expr;
      return e;
    case LETEXP:
      e->kind = A_LETEXP;
      e->u.letExp = (A_letExp)expr;
      return e;
    case OCLMESSAGEEXP:
      e->kind = A_OCLMESSAGEEXP;
      e->u.oclMessageExp = (A_oclMessageExp)expr;
      return e;
    case IFEXP:
      e->kind = A_IFEXP;
      e->u.ifExp = (A_ifExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}


A_variableExp A_VariableExp(A_name n)
{
  A_variableExp e = (A_variableExp)checked_malloc(sizeof(*e));
  e->type = VARIABLEEXP;
  e->name = n;
  return e;
}

A_name
A_Name(string s, A_name n)
{
  A_name e = (A_name)checked_malloc(sizeof(*e));
  e->type = NAME;
  e->str_value = s;
  e->next = n;
  return e;
}
  
A_literalExp
A_LiteralExp(void *expr, NodeType type)
{
  A_literalExp e = (A_literalExp)checked_malloc(sizeof(*e));
  e->type = LITERALEXP;
  switch (type)
    {
    case ENUMLITERALEXP:
      e->kind = A_ENUMLITERALEXP;
      e->u.enumLiteralExp = (A_enumLiteralExp)expr;
      return e;
    case COLLECTIONLITERALEXP:
      e->kind = A_COLLECTIONLITERALEXP;
      e->u.collectionLiteralExp = (A_collectionLiteralExp)expr;
      return e;
    case TUPLELITERALEXP:
      e->kind = A_TUPLELITERALEXP;
      e->u.tupleLiteralExp = (A_tupleLiteralExp)expr;
      return e;
    case NUMERICLITERALEXP:
      e->kind = A_NUMERICLITERALEXP;
      e->u.numericLiteralExp = (A_numericLiteralExp)expr;
      return e;
    case STRINGLITERALEXP:
      e->kind = A_STRINGLITERALEXP;
      e->u.stringLiteralExp = (A_stringLiteralExp)expr;
      return e;
    case BOOLEANLITERALEXP:
      e->kind = A_BOOLEANLITERALEXP;
      e->u.booleanLiteralExp = (A_booleanLiteralExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}


A_enumLiteralExp
A_EnumLiteralExp(A_name name, string s)
{
  A_enumLiteralExp e = (A_enumLiteralExp)checked_malloc(sizeof(*e));
  e->type = ENUMLITERALEXP;
  e->enumeration = name;
  e->value = s;
  return e;
} 

A_collectionLiteralExp
A_CollectionLiteralExp(A_collectionTypeIdentifier type, A_collectionItem items)
{
  A_collectionLiteralExp e = (A_collectionLiteralExp)checked_malloc(sizeof(*e));
  e->type = COLLECTIONLITERALEXP;
  e->collectionType = type;
  e->collectionItem = items;
  return e;
}

A_collectionItem
A_CollectionItem_link(A_collectionItem h, A_collectionItem t)
{
  h->next = t;
  return h;
}

A_collectionItem
A_CollectionItem(void *item, NodeType type)
{
  A_collectionItem e = (A_collectionItem)checked_malloc(sizeof(*e));
  e->type = COLLECTIONITEM;
  switch (type)
    {
    case COLLECTIONRANGE:
      e->kind = A_COLLECTIONRANGE;
      e->u.range = (A_collectionRange)item;
      e->next = NULL;
      return e;
    case OCLEXPRESSION:
      e->kind = A_INDIVIDUALITEM;
      e->u.individualItem = (A_oclExpression)item;
      e->next = NULL;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_collectionRange
A_CollectionRange(A_oclExpression f, A_oclExpression l)
{
  A_collectionRange e = (A_collectionRange)checked_malloc(sizeof(*e));
  e->type = COLLECTIONRANGE;
  e->first = f;
  e->last = l;
  return e;
}

A_numericLiteralExp
A_NumericLiteralExp(void *expr, NodeType type)
{
  A_numericLiteralExp e = (A_numericLiteralExp)checked_malloc(sizeof(*e));
  e->type = NUMERICLITERALEXP;
  switch (type)
    {
    case INTEGERLITERALEXP:
      e->kind = A_INTEGERLITERALEXP;
      e->u.integerLiteralExp = (A_integerLiteralExp)expr;
      return e;
    case REALLITERALEXP:
      e->kind = A_REALLITERALEXP;
      e->u.realLiteralExp = (A_realLiteralExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_integerLiteralExp
A_IntegerLiteralExp(int v)
{
  A_integerLiteralExp e = (A_integerLiteralExp)checked_malloc(sizeof(*e));
  e->type = INTEGERLITERALEXP;
  e->value = v;
  return e;
}

A_realLiteralExp
A_RealLiteralExp(double v)
{
  A_realLiteralExp e = (A_realLiteralExp)checked_malloc(sizeof(*e));
  e->type = REALLITERALEXP;
  e->value = v;
  return e;
}

A_stringLiteralExp
A_StringLiteralExp(string v)
{
  A_stringLiteralExp e = (A_stringLiteralExp)checked_malloc(sizeof(*e));
  e->type = STRINGLITERALEXP;
  e->value = v;
  return e;
}

A_booleanLiteralExp
A_BooleanLiteralExp(bool v)
{
  A_booleanLiteralExp e = (A_booleanLiteralExp)checked_malloc(sizeof(*e));
  e->type = BOOLEANLITERALEXP;
  e->value = v;
  return e;
}

A_tupleLiteralExp
A_TupleLiteralExp(A_tupleLiteralExpParts p)
{
  A_tupleLiteralExp e = (A_tupleLiteralExp)checked_malloc(sizeof(*e));
  e->type = TUPLELITERALEXP;
  e->tupleLiteralExpParts = p;
  return e;
}

A_tupleLiteralExpParts
A_TupleLiteralExpParts(A_variableDeclaration h, A_tupleLiteralExpParts t)
{
  A_tupleLiteralExpParts e = (A_tupleLiteralExpParts)checked_malloc(sizeof(*e));
  e->type = TUPLELITERALEXPPARTS;
  e->head = h;
  e->tail = t;
  return e;
}

A_propertyCallExp
A_PropertyCallExp(void *expr, NodeType type)
{
  A_propertyCallExp e = (A_propertyCallExp)checked_malloc(sizeof(*e));
  e->type = PROPERTYCALLEXP;
  switch (type)
    {
    case MODELPROPERTYCALLEXP:
      e->kind = A_MODELPROPERTYCALLEXP;
      e->u.modelPropertyCallExp = (A_modelPropertyCallExp)expr;
      return e;
    case LOOPEXP:
      e->kind = A_LOOPEXP;
      e->u.loopExp = (A_loopExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_loopExp
A_LoopExp(void *expr, NodeType type)
{
  A_loopExp e = (A_loopExp)checked_malloc(sizeof(*e));
  e->type = LOOPEXP;
  switch (type)
    {
    case ITERATOREXP:
      e->kind = A_ITERATOREXP;
      e->u.iteratorExp = (A_iteratorExp)expr;
      return e;
    case ITERATEEXP:
      e->kind = A_ITERATEEXP;
      e->u.iterateExp = (A_iterateExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_iteratorExp
A_IteratorExp(A_oclExpression s, A_collectionOp op,
	      A_variableDeclaration i, A_oclExpression b)
{
  A_iteratorExp e = (A_iteratorExp)checked_malloc(sizeof(*e));
  e->type = ITERATOREXP;
  e->source = s;
  e->body = b;
  e->op = op;
  e->iterator = i;
  return e;
}

A_iterateExp
A_IterateExp(A_oclExpression s, A_variableDeclaration i,
	      A_variableDeclaration r, A_oclExpression b)
{
  A_iterateExp e = (A_iterateExp)checked_malloc(sizeof(*e));
  e->type = ITERATEEXP;
  e->source = s;
  e->body = b;
  e->result = r;
  e->iterator = i;
  return e;
}

A_variableDeclaration
A_VariableDeclaration(string var, A_name type, A_oclExpression expr)
{
  A_variableDeclaration e = (A_variableDeclaration)checked_malloc(sizeof(*e));
  e->type = VARIABLEDECLARATION;
  e->varName = var;
  e->varType = type;
  e->initExpression = expr;
  return e;
}
    
A_modelPropertyCallExp
A_ModelPropertyCallExp(void *expr, NodeType type)
{
  A_modelPropertyCallExp e = (A_modelPropertyCallExp)checked_malloc(sizeof(*e));
  e->type = MODELPROPERTYCALLEXP;
  switch (type)
    {
    case OPERATIONCALLEXP:
      e->kind = A_OPERATIONCALLEXP;
      e->u.operationCallExp = (A_operationCallExp)expr;
      return e;
    case ATTRORNAVCALLEXP:
      e->kind = A_ATTRORNAVCALLEXP;
      e->u.attrOrNavCallExp = (A_attrOrNavCallExp)expr;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_operationCallExp
A_OperationCallExp(A_oclExpression s, string op, bool b, A_arguments args)
{
  A_operationCallExp e = (A_operationCallExp)checked_malloc(sizeof(*e));
  e->type = OPERATIONCALLEXP;
  e->source = s;
  e->op = op;
  e->timeExpr = b;
  e->arguments = args;
  return e;
}


A_arguments
A_Arguments(A_oclExpression h, A_arguments t)
{
  A_arguments e = (A_arguments)checked_malloc(sizeof(*e));
  e->type = ARGUMENTS;
  e->head = h;
  e->tail = t;
  return e;
}

A_attrOrNavCallExp
A_AttrOrNavCallExp(A_oclExpression src,string attrOrAss,bool b)
{
  A_attrOrNavCallExp e = (A_attrOrNavCallExp)checked_malloc(sizeof(*e));
  e->type = ATTRORNAVCALLEXP;
  e->source = src;
  e->attrOrAss = attrOrAss;
  e->timeExpr = b;
  return e;
}

A_letExp
A_LetExp(A_variables vars, A_oclExpression expr)
{
  A_letExp e = (A_letExp)checked_malloc(sizeof(*e));
  e->type = LETEXP;
  e->vars = vars;
  e->letExpSub = expr;
  return e;
}

A_variables
A_Variables(A_variableDeclaration h, A_variables t)
{
  A_variables e = (A_variables)checked_malloc(sizeof(*e));
  e->type = VARIABLES;
  e->head = h;
  e->tail = t;
  return e;
}

A_oclMessageExp
A_OclMessageExp(A_oclExpression target,string actionOrSignal,A_oclMessageArg arg)
{
  A_oclMessageExp e = (A_oclMessageExp)checked_malloc(sizeof(*e));
  e->type = OCLMESSAGEEXP;
  e->target = target;
  e->actionOrSignal = actionOrSignal;
  e->oclMessageArg = arg;
  return e;
}

A_oclMessageArg
A_OclMessageArg_link(A_oclMessageArg h, A_oclMessageArg t)
{
  h->next = t;
  return h;
}

A_oclMessageArg
A_OclMessageArg(void *expr, NodeType type)
{
  A_oclMessageArg e = (A_oclMessageArg)checked_malloc(sizeof(*e));
  e->type = OCLMESSAGEARG;
  switch (type)
    {
    case UNSPECIFIEDVALUEEXP: 
      e->kind = A_UNSPECIFIEDVALUEEXP;
      e->u.unspecified = (A_unspecifiedValueExp)expr;
      e->next = NULL;
      return e;
    case OCLEXPRESSION:
      e->kind = A_OCLEXPRESSION;
      e->u.expression = (A_oclExpression)expr;
      e->next = NULL;
      return e;
    default:
      fprintf(stderr, "Unknown NodeType\n");
      exit(1);
    }
}

A_unspecifiedValueExp
A_UnspecifiedValueExp(A_name type)
{
  A_unspecifiedValueExp e = (A_unspecifiedValueExp)checked_malloc(sizeof(*e));
  e->type = UNSPECIFIEDVALUEEXP;
  e->varType = type;
  return e;
}

A_ifExp
A_IfExp(A_oclExpression cond,A_oclExpression thenExpr,A_oclExpression elseExpr)
{
  A_ifExp e = (A_ifExp)checked_malloc(sizeof(*e));
  e->type = IFEXP;
  e->condition = cond;
  e->thenExpression = thenExpr;
  e->elseExpression = elseExpr;
  return e;
}


/*************************************************************************
 * Functions to traverse the abstract syntax C-tree
 * to creat an AST in ATOM3
 *************************************************************************/
void
traverse(void *node, int level)
{
  float coord_x = X_INIT + X_OFFSET * level;
  char varName[255];
  NodeType * nt;

  A_oclExpression oclExpression;
  A_variableExp variableExp;
  A_name name;
  A_propertyCallExp propertyCallExp;
  A_modelPropertyCallExp modelPropertyCallExp;
  A_literalExp literalExp;
  A_attrOrNavCallExp attrOrNavCallExp;
  A_operationCallExp operationCallExp;
  A_arguments arguments;
  A_stringLiteralExp stringLiteralExp;

  if (node == NULL)
    return;
  
  nt = node;
  switch (*nt)
    {
    case OCLEXPRESSION:
      oclExpression = (A_oclExpression)node;
      oclExpression->obj_ind = obj_ind;
      switch (oclExpression->kind)
	{
	case A_PROPERTYCALLEXP:
	  traverse(oclExpression->u.propertyCallExp, level);
	  break;
	case A_VARIABLEEXP:
	  traverse(oclExpression->u.variableExp, level);
	  break;;
	case A_LITERALEXP:
	  traverse(oclExpression->u.literalExp, level);
	  break;;
	case A_LETEXP:
	  traverse(oclExpression->u.letExp, level);
	  break;;
	case A_OCLMESSAGEEXP:
	  traverse(oclExpression->u.oclMessageExp, level);
	  break;;
	case A_IFEXP:
	  traverse(oclExpression->u.ifExp, level);
	  break;
	default:
	  fprintf(stderr, "Error in switch(oclExpression->kind)\n");
	  break;
	}
      break;
    case VARIABLEEXP:
      variableExp = (A_variableExp)node;
      variableExp->obj_ind = obj_ind; /* this must be done before the corresponding 
			       * entity get created */
      creat_VariableExp(coord_x, coord_y);
      traverse(variableExp->name, level+1);
      /* creat the relation obj */
      creat_referredVariable(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(variableExp->obj_ind, obj_ind-1, variableExp->name->obj_ind);
      return;
    case NAME:
      name = (A_name)node;
      name->obj_ind = obj_ind;
      strcpy(varName,name->str_value);
      while (name->next != NULL)
	{
	  name = name->next;
	  strcpy(varName, "::");
	  strcpy(varName, name->str_value);
	}
      creat_VariableDeclaration(varName, my_default_type, coord_x, coord_y);
      coord_y += Y_OFFSET;
      return;
    case PROPERTYCALLEXP:
      propertyCallExp = (A_propertyCallExp)node;
      propertyCallExp->obj_ind = obj_ind;
      switch (propertyCallExp->kind)
	{
	case A_MODELPROPERTYCALLEXP:
	  traverse(propertyCallExp->u.modelPropertyCallExp, level);
	  break;
	case A_LOOPEXP:
	  traverse(propertyCallExp->u.loopExp, level);
	  break;
	default:
	  fprintf(stderr, "Error\n");
	  break;
	}
      return;
    case MODELPROPERTYCALLEXP:
      modelPropertyCallExp = (A_modelPropertyCallExp)node;
      modelPropertyCallExp->obj_ind = obj_ind;
      switch (modelPropertyCallExp->kind)
	{
	case A_OPERATIONCALLEXP:
	  traverse(modelPropertyCallExp->u.operationCallExp, level);
	  break;
	case A_ATTRORNAVCALLEXP:
	  traverse(modelPropertyCallExp->u.attrOrNavCallExp, level);
	  break;
	default:
	  fprintf(stderr, "Error\n");
	  break;
	}
      return;
    case LITERALEXP:
      literalExp = (A_literalExp)node;
      literalExp->obj_ind = obj_ind;
      switch (literalExp->kind)
	{
	case A_ENUMLITERALEXP:
	  traverse(literalExp->u.enumLiteralExp, level);
	  break;
	case A_COLLECTIONLITERALEXP:
	  traverse(literalExp->u.collectionLiteralExp, level);
	  break;
	case A_TUPLELITERALEXP:
	  traverse(literalExp->u.tupleLiteralExp, level);
	  break;
	case A_NUMERICLITERALEXP:
	  traverse(literalExp->u.numericLiteralExp, level);
	  break;
	case A_STRINGLITERALEXP:
	  traverse(literalExp->u.stringLiteralExp, level);
	  break;
	case A_BOOLEANLITERALEXP:
	  traverse(literalExp->u.booleanLiteralExp, level);
	  break;
	default:
	  fprintf(stderr, "Error\n");
	  break;
	}
      return;
    case ATTRORNAVCALLEXP:
      attrOrNavCallExp = (A_attrOrNavCallExp)node;
      attrOrNavCallExp->obj_ind = obj_ind;
      /* We will creat an AttributeCallExp instance in ATOM3 for either
       * an AttributeCallExp or a NavigationCallExp
       * They will be sperated later with graph grammar in ATOM3
       */
      creat_AttributeCallExp(coord_x, coord_y);
      traverse(attrOrNavCallExp->source,level+1);
      creat_propertyCallSource(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(attrOrNavCallExp->obj_ind, obj_ind - 1, attrOrNavCallExp->source->obj_ind);
      creat_Attribute(attrOrNavCallExp->attrOrAss, coord_x + X_OFFSET, coord_y);
      creat_referredAttribute(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(attrOrNavCallExp->obj_ind, obj_ind - 1, obj_ind - 2);
      coord_y += Y_OFFSET;
      return ;
    case OPERATIONCALLEXP:
      operationCallExp = (A_operationCallExp)node;
      operationCallExp->obj_ind = obj_ind;
      creat_OperationCallExp(coord_x, coord_y);
      traverse(operationCallExp->source, level+1);
      creat_propertyCallSource(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(operationCallExp->obj_ind, obj_ind - 1, operationCallExp->source->obj_ind);
      creat_Operation(operationCallExp->op, coord_x + X_OFFSET, coord_y);
      creat_referredOperation(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(operationCallExp->obj_ind, obj_ind - 1, obj_ind - 2);
      coord_y += Y_OFFSET;
      if (operationCallExp->arguments == NULL)
	return;
      traverse(operationCallExp->arguments, level+1);
      creat_operationCallArguments(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(operationCallExp->obj_ind, obj_ind - 1, operationCallExp->arguments->obj_ind);
      return;
    case ARGUMENTS: /* non empty arguments */
      arguments = (A_arguments)node;
      arguments->obj_ind = obj_ind;
      creat_Arguments(coord_x, coord_y);
      traverse(arguments->head, level+1);
      creat_first(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(arguments->obj_ind, obj_ind - 1, arguments->head->obj_ind);
      if (arguments->tail == NULL)
	return;
      traverse(arguments->tail,level+1);
      creat_tail(coord_x + X_OFFSET2, coord_y - Y_OFFSET);
      connect(arguments->obj_ind, obj_ind - 1, arguments->tail->obj_ind);
      return;
    case STRINGLITERALEXP:
      stringLiteralExp = (A_stringLiteralExp)node;
      stringLiteralExp->obj_ind = obj_ind;
      creat_StringLiteralExp(stringLiteralExp->value, coord_x, coord_y);
      coord_y += Y_OFFSET;
      return;
    default:
      fprintf(stderr, "Unrecongnized node type\n");
      return;
    }
}
      
      
      
      
      
      
      

