/* 
 * absyn.h
 * This file will define all the constructs and function prototypes
 * for building the abstract syntax tree
 * Please note that the abstract syntax tree is not an exact instance
 * of the abstract syntax defined in the OCL2.0 specification, it is 
 * just a syntax tree build from the concrete syntax by discarding 
 * the unnecessary info
 */

#define Y_INIT 50.0
#define Y_OFFSET 100.0
#define X_INIT 50.0
#define X_OFFSET 200.0
#define X_OFFSET2 150.0


typedef struct A_oclExpression_ *A_oclExpression;
typedef struct A_propertyCallExp_ *A_propertyCallExp;
typedef struct A_variableExp_ *A_variableExp;
typedef struct A_literalExp_ *A_literalExp;
typedef struct A_letExp_ *A_letExp;
typedef struct A_oclMessageExp_ *A_oclMessageExp;
typedef struct A_ifExp_ *A_ifExp;

typedef struct A_name_ *A_name;

typedef struct A_enumLiteralExp_ *A_enumLiteralExp;
typedef struct A_collectionLiteralExp_ *A_collectionLiteralExp;
typedef struct A_tupleLiteralExp_ *A_tupleLiteralExp;
typedef struct A_numericLiteralExp_ *A_numericLiteralExp;
typedef struct A_stringLiteralExp_ *A_stringLiteralExp;
typedef struct A_booleanLiteralExp_ *A_booleanLiteralExp;

typedef enum{A_SET, A_BAG, A_SEQUENCE} A_collectionTypeIdentifier;
typedef struct A_collectionItem_ *A_collectionItem;
typedef struct A_collectionRange_ *A_collectionRange;

typedef struct A_integerLiteralExp_ *A_integerLiteralExp;
typedef struct A_realLiteralExp_ *A_realLiteralExp;

typedef struct A_tupleLiteralExpParts_ *A_tupleLiteralExpParts;

typedef struct A_modelPropertyCallExp_ *A_modelPropertyCallExp;
typedef struct A_loopExp_ *A_loopExp;

typedef struct A_iteratorExp_ *A_iteratorExp;
typedef struct A_iterateExp_ *A_iterateExp;
typedef enum{  A_SELECT, A_COLLECT, A_REJECT, A_FORALL, A_EXISTS,
	       A_ISUNIQUE, A_SIZE, A_ISEMPTY, A_NOTEMPTY,
		 A_COUNT, A_UNION, A_DIFFERENCE } A_collectionOp;

typedef struct A_variableDeclaration_ *A_variableDeclaration;

typedef struct A_operationCallExp_ *A_operationCallExp;
typedef struct A_attrOrNavCallExp_ *A_attrOrNavCallExp;

typedef struct A_arguments_ *A_arguments; 

typedef struct A_variables_ *A_variables;

typedef struct A_oclMessageArg_ *A_oclMessageArg;
typedef struct A_unspecifiedValueExp_ *A_unspecifiedValueExp;
 
typedef
enum { OCLEXPRESSION=0, PROPERTYCALLEXP, VARIABLEEXP, LITERALEXP, LETEXP,
       OCLMESSAGEEXP, IFEXP, NAME, ENUMLITERALEXP, COLLECTIONLITERALEXP,
       TUPLELITERALEXP, NUMERICLITERALEXP, STRINGLITERALEXP,
       BOOLEANLITERALEXP, COLLECTIONTYPEIDENTIFIER, COLLECTIONITEM,
       COLLECTIONRANGE, INTEGERLITERALEXP, REALLITERALEXP, 
       TUPLELITERALEXPPARTS, VARIABLEDECLARATION, MODELPROPERTYCALLEXP,
       LOOPEXP, ITERATOREXP, ITERATEEXP, COLLECTIONOP, OPERATIONCALLEXP,
       ATTRORNAVCALLEXP, VARIABLES, OCLMESSAGEARG, ARGUMENTS,
       UNSPECIFIEDVALUEEXP} NodeType;

       

/*************************************************************************
 * The actual struct definitions
 *************************************************************************/
struct A_oclExpression_ 
{
  NodeType type;
  enum { A_PROPERTYCALLEXP,
	 A_VARIABLEEXP,
	 A_LITERALEXP,
	 A_LETEXP,
	 A_OCLMESSAGEEXP,
	 A_IFEXP
       } kind;
  union{ 
    A_propertyCallExp propertyCallExp;
    A_variableExp variableExp;
    A_literalExp literalExp;
    A_letExp letExp;
    A_oclMessageExp oclMessageExp;
    A_ifExp ifExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_variableExp_
{
  NodeType type;
  A_name name;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_name_
{
  NodeType type;
  string str_value;
  A_name next;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_literalExp_
{
  NodeType type;
  enum {A_ENUMLITERALEXP, A_COLLECTIONLITERALEXP, A_TUPLELITERALEXP,
	  A_NUMERICLITERALEXP, A_STRINGLITERALEXP, A_BOOLEANLITERALEXP } kind;
  union{
    A_enumLiteralExp enumLiteralExp;
    A_collectionLiteralExp collectionLiteralExp;
    A_tupleLiteralExp tupleLiteralExp;
    A_numericLiteralExp numericLiteralExp;
    A_stringLiteralExp stringLiteralExp;
    A_booleanLiteralExp booleanLiteralExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_enumLiteralExp_
{
  NodeType type;
  A_name enumeration;
  string value;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_collectionLiteralExp_
{
  NodeType type;
  A_collectionTypeIdentifier collectionType;
  A_collectionItem collectionItem;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_collectionItem_
{
  NodeType type;
  enum{A_COLLECTIONRANGE, A_INDIVIDUALITEM} kind;
  union{
    A_collectionRange range;
    A_oclExpression individualItem;
  } u;
  A_collectionItem next;
};

struct A_collectionRange_
{
  NodeType type;
  A_oclExpression first;
  A_oclExpression last;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_numericLiteralExp_
{
  NodeType type;
  enum{ A_INTEGERLITERALEXP, A_REALLITERALEXP } kind;
  union{
    A_integerLiteralExp integerLiteralExp;
    A_realLiteralExp realLiteralExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_integerLiteralExp_
{
  NodeType type;
  int value;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_realLiteralExp_
{
  NodeType type;
  double value;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_stringLiteralExp_
{
  NodeType type;
  string value;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_booleanLiteralExp_
{
  NodeType type;
  bool value;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_tupleLiteralExp_
{
  NodeType type;
  A_tupleLiteralExpParts tupleLiteralExpParts;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_tupleLiteralExpParts_
{
  NodeType type;
  A_variableDeclaration head;
  A_tupleLiteralExpParts tail;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_propertyCallExp_
{
  NodeType type;
  enum{ A_MODELPROPERTYCALLEXP, A_LOOPEXP } kind;
  union{
    A_modelPropertyCallExp modelPropertyCallExp;
    A_loopExp loopExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_loopExp_
{
  NodeType type;
  enum{ A_ITERATOREXP, A_ITERATEEXP } kind;
  union{
    A_iteratorExp iteratorExp;
    A_iterateExp iterateExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_iteratorExp_
{
  NodeType type;
  A_oclExpression source;
  A_oclExpression body;
  A_variableDeclaration iterator;
  A_collectionOp op;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_iterateExp_
{
  NodeType type;

  A_oclExpression source;
  A_oclExpression body;
  A_variableDeclaration iterator;
  A_variableDeclaration result;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_variableDeclaration_
{
  NodeType type;
  string varName;
  A_name varType;
  A_oclExpression initExpression;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};


struct A_modelPropertyCallExp_
{
  NodeType type;
  enum{ A_OPERATIONCALLEXP, A_ATTRORNAVCALLEXP } kind;
  union{
    A_operationCallExp operationCallExp;
    A_attrOrNavCallExp attrOrNavCallExp;
  } u;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

/*
 * since the operation includes those from the UML model
 * its representation should be a string rather than an enum
 */
struct A_operationCallExp_
{
  NodeType type;
  A_oclExpression source;
  A_arguments arguments;
  string op;
  bool timeExpr;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_attrOrNavCallExp_ 
{
  NodeType type;
  A_oclExpression source;
  string attrOrAss;
  bool timeExpr;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_arguments_
{
  NodeType type;
  A_oclExpression head;
  A_arguments tail;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_letExp_
{
  NodeType type;
  A_variables vars;
  A_oclExpression letExpSub;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_variables_
{
  NodeType type;
  A_variableDeclaration head;
  A_variables tail;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_oclMessageExp_
{
  NodeType type;
  A_oclExpression target;
  string actionOrSignal;
  A_oclMessageArg oclMessageArg;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_oclMessageArg_
{
  NodeType type;
  enum{ A_UNSPECIFIEDVALUEEXP, A_OCLEXPRESSION } kind;
  union{
    A_unspecifiedValueExp unspecified;
    A_oclExpression expression;
  } u;
  A_oclMessageArg next;
};

struct A_unspecifiedValueExp_
{
  NodeType type;
  A_name varType;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};

struct A_ifExp_
{
  NodeType type;
  A_oclExpression condition;
  A_oclExpression thenExpression;
  A_oclExpression elseExpression;
  /* The following attributes used for building up a AST in Atom3 env */
  int obj_ind;
  int level;
};



/*************************************************************************
 * Constructor prototypes
 *************************************************************************/
A_oclExpression
A_OclExpression(void *expr, NodeType type);


A_variableExp
A_VariableExp(A_name n);

A_name
A_Name(string s, A_name n);
  
A_literalExp
A_LiteralExp(void *expr, NodeType type);

A_enumLiteralExp
A_EnumLiteralExp(A_name name, string s);

A_collectionLiteralExp
A_CollectionLiteralExp(A_collectionTypeIdentifier type, A_collectionItem items);
A_collectionItem
A_CollectionItem_link(A_collectionItem h, A_collectionItem t);

A_collectionItem
A_CollectionItem(void *item, NodeType type);

A_collectionRange
A_CollectionRange(A_oclExpression f, A_oclExpression l);

A_numericLiteralExp
A_NumericLiteralExp(void *expr, NodeType type);

A_integerLiteralExp
A_IntegerLiteralExp(int v);

A_realLiteralExp
A_RealLiteralExp(double v);

A_stringLiteralExp
A_StringLiteralExp(string v);

A_booleanLiteralExp
A_BooleanLiteralExp(bool v);

A_tupleLiteralExp
A_TupleLiteralExp(A_tupleLiteralExpParts p);

A_tupleLiteralExpParts
A_TupleLiteralExpParts(A_variableDeclaration h, A_tupleLiteralExpParts t);

A_propertyCallExp
A_PropertyCallExp(void *expr, NodeType type);

A_loopExp
A_LoopExp(void *expr, NodeType type);


A_iteratorExp
A_IteratorExp(A_oclExpression s, A_collectionOp op,
	      A_variableDeclaration i, A_oclExpression b);

A_iterateExp
A_IterateExp(A_oclExpression s, A_variableDeclaration i,
	     A_variableDeclaration r, A_oclExpression b);

A_variableDeclaration
A_VariableDeclaration(string var, A_name type, A_oclExpression expr);
    
A_modelPropertyCallExp
A_ModelPropertyCallExp(void *expr, NodeType type);

A_operationCallExp
A_OperationCallExp(A_oclExpression s, string op, bool b, A_arguments args);


A_arguments
A_Arguments(A_oclExpression h, A_arguments t);

A_attrOrNavCallExp
A_AttrOrNavCallExp(A_oclExpression src,string attrOrAss,bool b);

A_letExp
A_LetExp(A_variables vars, A_oclExpression expr);

A_variables
A_Variables(A_variableDeclaration h, A_variables t);

A_oclMessageExp
A_OclMessageExp(A_oclExpression target,string actionOrSignal,A_oclMessageArg arg);

A_oclMessageArg
A_OclMessageArg_link(A_oclMessageArg h, A_oclMessageArg t);

A_oclMessageArg
A_OclMessageArg(void *expr, NodeType type);

A_unspecifiedValueExp
A_UnspecifiedValueExp(A_name type);

A_ifExp
A_IfExp(A_oclExpression cond,A_oclExpression thenExpr,A_oclExpression elseExpr);

void
traverse(void *node, int level);
