/*
 * This file contains subroutines to create python codes to 
 * create AST of OCL expression
 * 
 * We will have a general subroutine to creat relation object
 * And we will also have a general subroutine to creat different entity objects
 * Since they have different attributes, we will have a pointer to a linked list
 * of attributes as an argument
 *
 * Problems:
 * 1.How to determine the coordinates of the objects and their connections
 * 2.Everything is hard-coded, if the meta-model changes, all the functions
 *   may have to be changed.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "creat_python.h"

char *strdup(const char *s);

extern FILE *AST_out;


int obj_ind;



/*************************************************************************************
 * Creat python codes for all the neccessary imports classes
 *************************************************************************************/
void
creat_imports()
{
  if (AST_out == NULL)
    {
      fprintf(stderr, "The file handler passed to the function is not valid\n");
      exit(1);
    }
  fprintf(AST_out, "from graph_ASG_ERmetaMetaModel import *\n");
  fprintf(AST_out, "from stickylink import *\n");
  fprintf(AST_out, "from widthXfillXdecoration import *\n");
  fprintf(AST_out, "from ASG_ERmetaMetaModel import *\n");
  fprintf(AST_out, "from ERentity import *\n");
  fprintf(AST_out, "from ERrelationship import *\n");
  fprintf(AST_out, "from propertyCallSource import *\n");
  fprintf(AST_out, "from referredOperation import *\n");
  fprintf(AST_out, "from IteratorExp import *\n");
  fprintf(AST_out, "from iterator import *\n");
  fprintf(AST_out, "from OclExpression import *\n");
  fprintf(AST_out, "from Arguments import *\n");
  fprintf(AST_out, "from referredVariable import *\n");
  fprintf(AST_out, "from referredAttribute import *\n");
  fprintf(AST_out, "from operationCallArguments import *\n");
  fprintf(AST_out, "from ASG_P1_Formalism import *\n");
  fprintf(AST_out, "from loopBody import *\n");
  fprintf(AST_out, "from VariableDeclaration import *\n");
  fprintf(AST_out, "from OperationCallExp import *\n");
  fprintf(AST_out, "from PropertyCallExp import *\n");
  fprintf(AST_out, "from StringLiteralExp import *\n");
  fprintf(AST_out, "from Attribute import *\n");
  fprintf(AST_out, "from VariableExp import *\n");
  fprintf(AST_out, "from ModelPropertyCall import *\n");
  fprintf(AST_out, "from tail import *\n");
  fprintf(AST_out, "from AttributeCallExp import *\n");
  fprintf(AST_out, "from LiteralExp import *\n");
  fprintf(AST_out, "from Operation import *\n");
  fprintf(AST_out, "from initExpr import *\n");
  fprintf(AST_out, "from LoopExp import *\n");
  fprintf(AST_out, "from first import *\n");
  fprintf(AST_out, "from ATOM3Enum import *\n");
  fprintf(AST_out, "from ATOM3List import *\n");
  fprintf(AST_out, "from ATOM3Float import *\n");
  fprintf(AST_out, "from ATOM3Integer import *\n");
  fprintf(AST_out, "from ATOM3Attribute import *\n");
  fprintf(AST_out, "from ATOM3Constraint import *\n");
  fprintf(AST_out, "from ATOM3String import *\n");
  fprintf(AST_out, "from ATOM3BottomType import *\n");
  fprintf(AST_out, "from ATOM3Boolean import *\n");
  fprintf(AST_out, "from ATOM3Appearance import *\n");
  fprintf(AST_out, "from ATOM3Link import *\n");
  fprintf(AST_out, "\n\n");
}


/************************************************************************************** 
 * Creat the initial environment of the function
 **************************************************************************************/
void
creat_init(char *fn_name)
{
  fprintf(AST_out, "def %s(self, rootNode):\n", fn_name);
  fprintf(AST_out, "\trootNode.attributes.setActionFlags([1,1,1,0])\n");
  fprintf(AST_out, "\tlcobj1 = []\n");
  fprintf(AST_out, "\trootNode.attributes.setValue(lcobj1)\n");
  fprintf(AST_out, "\trootNode.name.setValue('ERModel')\n");
  fprintf(AST_out, "\trootNode.constraints.setActionFlags([1,1,1,0])\n");
  fprintf(AST_out, "\tlcobj1 = []\n");
  fprintf(AST_out, "\trootNode.constraints.setValue(lcobj1)\n");
  fprintf(AST_out, "\trootNode.graphClass_ = graph_ASG_ERmetaMetaModel\n");
  fprintf(AST_out, "\n");
}


/**************************************************************************************
 * General function to creat a relation object
 * The object is not connected to anything else so far
 * x and y are the coordinates of the objects on the drawing area
 **************************************************************************************/
void
creat_relation(char *rel_name, char *role, char *role2, float x, float y)
{
  fprintf(AST_out, "\tself.globalPrecondition(rootNode)\n");
  fprintf(AST_out, "\tself.obj%d = %s(self)\n",obj_ind,rel_name);
  fprintf(AST_out, "\tself.obj%d.roleName.setValue('%s')\n",obj_ind,role);
  fprintf(AST_out, "\tself.obj%d.roleName2.setValue('%s')\n",obj_ind,role2);
  fprintf(AST_out, "\tself.obj%d.graphCalss_ = graph_%s\n",obj_ind,rel_name);
  fprintf(AST_out, "\tif self.genGraphics:\n");
  fprintf(AST_out, "\t\tfrom graph_%s import *\n",rel_name);
  fprintf(AST_out, "\t\tnew_obj = graph_%s(%f,%f,self.obj%d)\n",rel_name,x,y,obj_ind);
  fprintf(AST_out, "\t\tnew_obj.DrawObject(self.UMLmodel)\n");
  fprintf(AST_out, "\t\tself.UMLmodel.addtag_withtag(\"%s\",new_obj.tag)\n",rel_name);
  fprintf(AST_out, "\telse: new_obj = None\n");
  fprintf(AST_out, "\tself.obj%d.graphObject_ = new_obj\n",obj_ind);
  fprintf(AST_out, "\trootNode.addNode(self.obj%d)\n",obj_ind);
  fprintf(AST_out, "\tself.globalAndLocalPostcondition(self.obj%d,rootNode)\n",obj_ind);
  fprintf(AST_out, "\n");

  obj_ind++; /* increment the object index */
}


/************************************************************************************** 
 * Creat a propertyCallSource relation object--
 **************************************************************************************/
void
creat_propertyCallSource(float x, float y)
{
  fprintf(AST_out, "\t# creat a propertyCallSource relation obj\n");
  creat_relation("propertyCallSource","source","appliedProperty",x,y);
}


/************************************************************************************** 
 * Creat an operationCallArguments relation object--
 **************************************************************************************/
void
creat_operationCallArguments(float x, float y)
{
  fprintf(AST_out, "\t# creat an operationCallArguments relation obj\n");
  creat_relation("operationCallArguments","arguments",NULL,x,y);
}


/************************************************************************************** 
 * Creat a referredVariable relation object--
 **************************************************************************************/
void
creat_referredVariable(float x, float y)
{
  fprintf(AST_out, "\t# creat a referredVariable relation obj\n");
  creat_relation("referredVariable","referredVariable",NULL,x,y);
}


/************************************************************************************** 
 * Creat a referredAttribute relation object--
 **************************************************************************************/
void
creat_referredAttribute(float x, float y)
{
  fprintf(AST_out, "\t# creat a referredAttribute relation obj\n");
  creat_relation("referredAttribute","referredAttribute",NULL,x,y);
}


/************************************************************************************** 
 * Creat a referredOperation relation object--
 **************************************************************************************/
void
creat_referredOperation(float x, float y)
{
  fprintf(AST_out, "\t# creat a referredOperation relation obj\n");
  creat_relation("referredOperation","referredOperation",NULL,x,y);
}

/************************************************************************************** 
 * Creat a first relation object--
 **************************************************************************************/
void
creat_first(float x, float y)
{
  fprintf(AST_out, "\t# creat a first relation obj\n");
  creat_relation("first","head","parentOperation",x,y);
}

/************************************************************************************** 
 * Creat a first relation object--
 **************************************************************************************/
void
creat_tail(float x, float y)
{
  fprintf(AST_out, "\t# creat a tail relation obj\n");
  creat_relation("tail","tail",NULL,x,y);
}


/************************************************************************************* 
 * General function to creat an entity object 
 *************************************************************************************/
void 
creat_entity(char *entity_name, ATTRIB *attrib, float x, float y)
{
  ATTRIB *current;
  ATTRIB *tmp;

  fprintf(AST_out, "\tself.globalPrecondition(rootNode)\n");
  fprintf(AST_out, "\tself.obj%d = %s(self)\n",obj_ind,entity_name);

  /* initiate the attributes of the entity */
  current = attrib;
  while (current != NULL)
    {
      fprintf(AST_out, "\tself.obj%d.%s.setValue('%s')\n",obj_ind,current->name,current->value);
      tmp = current;
      current = current->next;
      free(tmp);
      tmp = NULL;
    }
 
  fprintf(AST_out, "\tself.obj%d.graphCalss_ = graph_%s\n",obj_ind,entity_name);
  fprintf(AST_out, "\tif self.genGraphics:\n");
  fprintf(AST_out, "\t\tfrom graph_%s import *\n",entity_name);
  fprintf(AST_out, "\t\tnew_obj = graph_%s(%f,%f,self.obj%d)\n",entity_name,x,y,obj_ind);
  fprintf(AST_out, "\t\tnew_obj.DrawObject(self.UMLmodel)\n");
  fprintf(AST_out, "\t\tself.UMLmodel.addtag_withtag(\"%s\",new_obj.tag)\n",entity_name);
  fprintf(AST_out, "\telse: new_obj = None\n");
  fprintf(AST_out, "\tself.obj%d.graphObject_ = new_obj\n",obj_ind);
  fprintf(AST_out, "\trootNode.addNode(self.obj%d)\n",obj_ind);
  fprintf(AST_out, "\tself.globalAndLocalPostcondition(self.obj%d,rootNode)\n",obj_ind);
  fprintf(AST_out, "\n");

  obj_ind++; /* increment the object index */
}

/**************************************************************************************
 * Creat an Arguments entity object-- 
 **************************************************************************************/
void
creat_Arguments(float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  assert(attrib);
  attrib->name = (char *)strdup("expType");
  attrib->value = (char *)strdup("Arguments");
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat an Arguments entity obj\n");
  creat_entity( "Arguments", attrib, x, y);
}

/**************************************************************************************
 * Creat an Operation entity object-- 
 **************************************************************************************/
void
creat_Operation(char *operator, float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("operator");
  attrib->value = (char *)strdup(operator);
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat an Operation entity obj\n");
  creat_entity( "Operation", attrib, x, y);
}

/*************************************************************************************
 * Creat an AttributeCallExp entity object-- 
 **************************************************************************************/
void
creat_AttributeCallExp(float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("expType");
  attrib->value = (char *)strdup("AttributeCallExp");
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat an AttributeCallExp entity obj\n");
  creat_entity( "AttributeCallExp", attrib, x, y);
}

/**************************************************************************************
 * Creat a VariableExp entity object-- 
 **************************************************************************************/
void
creat_VariableExp(float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("expType");
  attrib->value = (char *)strdup("VariableExp");
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat a VariableExp entity obj\n");
  creat_entity( "VariableExp", attrib, x, y);
}


/**************************************************************************************
 * Creat an Attribute entity object-- 
 **************************************************************************************/
void
creat_Attribute(char *attribute_name, float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("attributeName");
  attrib->value = (char *)strdup(attribute_name);
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat an Attribute entity obj\n");
  creat_entity( "Attribute", attrib, x, y);
}


/*************************************************************************************
 * Creat a StringLiteralExp entity object-- 
 *************************************************************************************/
void
creat_StringLiteralExp(char *string_value, float x, float y)
{
  ATTRIB *attrib;
  ATTRIB *tmp;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("expType");
  attrib->value = (char *)strdup("StringLiteralExp");
  attrib->next = NULL;
  tmp = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (tmp == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  tmp->name = (char *)strdup("value");
  tmp->value = (char *)strdup(string_value);
  tmp->next = attrib;
  attrib = tmp;
  fprintf(AST_out, "\t# creat a StringLiteralExp entity obj\n");
  creat_entity( "StringLiteralExp", attrib, x, y);
}


/**************************************************************************************
 * Creat a VariableDeclaration entity object-- 
 **************************************************************************************/
void
creat_VariableDeclaration(char *varName, char *type, float x, float y)
{
  ATTRIB *attrib;
  ATTRIB *tmp;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("type");
  attrib->value = (char *)strdup(type);
  attrib->next = NULL;
  tmp = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (tmp == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  tmp->name = (char *)strdup("varName");
  tmp->value = (char *)strdup(varName);
  tmp->next = attrib;
  attrib = tmp;
  fprintf(AST_out, "\t# creat a VariableDeclaration entity obj\n");
  creat_entity("VariableDeclaration", attrib, x, y);
}


/**************************************************************************************
 * Creat an OperationCallExp entity object-- 
 **************************************************************************************/
void
creat_OperationCallExp(float x, float y)
{
  ATTRIB *attrib;
  attrib = (ATTRIB *)malloc(sizeof(ATTRIB));
  if (attrib == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }
  attrib->name = (char *)strdup("expType");
  attrib->value = (char *)strdup("OperationCallExp");
  attrib->next = NULL;
  fprintf(AST_out, "\t# creat an OperationCallExp entity obj\n");
  creat_entity("OperationCallExp", attrib, x, y);
}

/**************************************************************************************
 * Connet two entity objects with a relation object
 **************************************************************************************/
void
connect(int i, int i2, int i3)
{
  fprintf(AST_out,"\tself.drawConnections((self.obj%d,self.obj%d),(self.obj%d,self.obj%d))\n",i, i2, i2, i3);
}


/**************************************************************************************
 * Creat the end part of the function
 **************************************************************************************/
void
creat_end(char *fn_name)
{
  fprintf(AST_out,"\nnewfunction = %s\n\n", fn_name);
  fprintf(AST_out,"loadedMMName = 'P1_Formalism'\n");
}


