%{
/* The first ocl Lexer */

#include "util.h"
#include "absyn.h"
#include "y.tab.h"
#include <stdlib.h>
#include <string.h>

extern int lineno;
extern void
warning(char *s, char *t);

extern char *strdup(const char *s);
%}

ws  [ \t\r]+
comment "--".*
qstring \"(\\\"|[^\n\"])*[\n\"]
dot "."
ddot ".."
dcolon  "::"
arrow "->"
atpre "@pre"
nl  \n

%%

{ws}  ;
{comment}  { printf("A line of comment was parsed\n"); }

  /* QSTRING
   * The quoted string should be in one line
   * The exact syntax of a String is undefined in UML1.4 and OCL2.0 for internalization
   */
{qstring}  { if(yytext[yyleng-1] != '"')
             {
               warning("Unterminated string", NULL);
               lineno++;
             }
             yylval.sval = (string)strdup(yytext);
             return QSTRING;
           }

  /* reserved keywords */
inv { return INV; }

pre { return PRE; }

post  { return POST; }

def { return DEF; }

Set { return SET; }

Bag { return BAG; }

Sequence  { return SEQUENCE; }

Tuple { return  TUPLE; }

iterate { return ITERATE; }

Let { return LET; }

in  { return IN; }

if  { return IF; }

then  { return THEN; }

else  { return ELSE; }

endif { return ENDIF; }

{dcolon}  { return DCOLON; }

{dot} {return DOT; }

{ddot}  { return DDOT; }

{arrow} { return ARROW; }

{atpre} { return ATPRE; }


  /* reserved relation operations */
"+" { return PLUS; }

"-" { return MINUS; }

"*" { return TIMES; }

"/" { return DIVIDES; }

"and" { return AND; }

"or"  { return OR; }

"xor" { return XOR; }

"implies" { return IMPLIES; }

"="   { return EQ; }

"<>"  { return NOTEQ; }

"<="  { return LTEQ; }

">="  { return GTEQ; }

"<"   { return LT; }

">"   { return GT; }


  /* reserved collection operations */
select  { return SELECT; }

collect { return COLLECT; }

reject  { return SYM_REJECT; }

forAll  { return FORALL; }

exists  { return EXISTS; }

isUnique  { return ISUNIQUE; }

size    { return SIZE; }

isEmpty { return ISEMPTY; }

notEmpty  { return NOTEMPTY; }

count   { return COUNT; }

union   { return UNION; }

difference  { return DIFFERENCE; }

  /* reserved negates operation */
not { return NOT; }



  /* numbers */
[0-9]+  { yylval.ival = atoi(yytext); return INTEGER; }

[0-9]+"."[0-9]* |
"."[0-9]+  |
[0-9]+[eE][-+]?[0-9]+ |
[0-9]+"."[0-9]*[eE][-+]?[0-9]+  |
"."[0-9]+[eE][-+]?[0-9]+  { yylval.dval = atof(yytext); return REAL; }


  /* boolean */
[Tt][Rr][Uu][Ee]  { return SYM_TRUE; }

[Ff][Aa][Ll][Ss][Ee]  { return SYM_FALSE; }


  /* name String
   * The exact syntax of a String is undefined in UML1.4 and OCL2.0 for internalization
   */
[A-Za-z][A-Za-z0-9_]* { yylval.sval = strdup(yytext); return NAMESTRING; }





{nl}  { lineno++;}
.     { return yytext[0]; }

%%

void
yyerror(char *msg)
{
  fprintf(stderr, "%d: %s at '%s'\n", lineno, msg, yytext);
}









